/************************************************************************ 
 * RSTP library - Rapid Spanning Tree (802.1t, 802.1w) 
 * Copyright (C) 2001-2003 Optical Access 
 * Author: Alex Rozin 
 * 
 * This file is part of RSTP library. 
 * 
 * RSTP library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU Lesser General Public License as published by the 
 * Free Software Foundation; version 2.1 
 * 
 * RSTP library is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser 
 * General Public License for more details. 
 * 
 * You should have received a copy of the GNU Lesser General Public License 
 * along with RSTP library; see the file COPYING.  If not, write to the Free 
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
 * 02111-1307, USA. 
 **********************************************************************/

/* Port Protocol Migration state machine : 17.26 */

#include "base.h"
#include "stpm.h"
#include "stp_to.h" /* for STP_OUT_get_port_oper_speed */

#define STATES { \
  CHOOSE(INIT),     \
  CHOOSE(SEND_RSTP),    \
  CHOOSE(SENDING_RSTP), \
  CHOOSE(SEND_STP), \
  CHOOSE(SENDING_STP),  \
}

#define GET_STATE_NAME STP_migrate_get_state_name
#include "choose.h"

#define MigrateTime 3		/* 17,16.4 */

void
STP_migrate_enter_state (STATE_MACH_T * this)
{
  register PORT_T *port = this->owner.port;

  switch (this->State) {
    case BEGIN:
    case INIT:
    port->initPm = True;
    port->mcheck = False;
    break;
    case SEND_RSTP:
    port->mdelayWhile = MigrateTime;
    port->mcheck = port->initPm = False;
    port->sendRSTP = True;
    break;
    case SENDING_RSTP:
    port->rcvdRSTP = port->rcvdSTP = False;
    break;
    case SEND_STP:
    port->mdelayWhile = MigrateTime;
    port->sendRSTP = False;
    port->initPm = False;
    break;
    case SENDING_STP:
    port->rcvdRSTP = port->rcvdSTP = False;
    break;
  }
}

/* version 0:stp, 2:rstp */
static long computeAdjPCost (PORT_T * port, int version)
{
  unsigned long spd;
  long lret;

  spd = STP_OUT_get_port_oper_speed (port->port_index);

  if (spd < 10L) { /* < 10Mb/s */
    lret = (version == 0)? 250 : 20000000;
  } else if (spd <= 10L) {
    lret = (version == 0)? 100 : 2000000;
  } else if (spd <= 20L) {
    lret = (version == 0)? 60 : 1800000;
  } else if (spd <= 30L) {
    lret = (version == 0)? 42 : 1600000;
  } else if (spd <= 40L) {
    lret = (version == 0)? 30 : 1400000;
  } else if (spd <= 100L) {
    lret = (version == 0)? 19 : 200000;
  } else if (spd <= 200L) {
    lret = (version == 0)? 16 : 180000;
  } else if (spd <= 300L) {
    lret = (version == 0)? 14 : 160000;
  } else if (spd <= 400L) {
    lret = (version == 0)? 12 : 140000;
  } else if (spd <= 1000L) {
    lret = (version == 0)? 4 : 20000;
  } else if (spd <= 2000L) {
    lret = (version == 0)? 3 : 18000;
  } else if (spd <= 10000L) {
    lret = (version == 0)? 1 : 2000;
  } else if (spd <= 100000L) {
    lret = (version == 0)? 1 : 200;
  } else if (spd <= 1000000L) {
    lret = (version == 0)? 1 : 20;
  } else if (spd <= 10000000L) {
    lret = (version == 0)? 1 : 2;
  } else {
    /* > Tb/s */
    lret = 1;
  }
  return lret;
}

Bool
STP_migrate_check_conditions (STATE_MACH_T * this)
{
  register PORT_T *port = this->owner.port;

  if ((!port->portEnabled && !port->initPm) || BEGIN == this->State)
    return STP_hop_2_state (this, INIT);

  switch (this->State) {
    case INIT:
    if (port->portEnabled) {
      return STP_hop_2_state (this, (port->owner->ForceVersion >= 2) ?
			      SEND_RSTP : SEND_STP);
    }
    break;
    case SEND_RSTP:
    return STP_hop_2_state (this, SENDING_RSTP);
    case SENDING_RSTP:
    if (port->mcheck)
      return STP_hop_2_state (this, SEND_RSTP);

    if (port->rcvdSTP && !port->rx_cfg_bpdu_cnt) {
      return STP_hop_2_state (this, SEND_STP);
    }

    if (port->mdelayWhile) {
      if (port->rcvdSTP || port->rcvdRSTP) {
	return STP_hop_2_state (this, SENDING_RSTP);
      }
    } else if (port->rcvdSTP) {
      port->adminPCost = computeAdjPCost (port, 0);
#ifdef STP_DBG
      if (port->pcost->debug) {
        stp_trace ("migrate adminPCost=%lu", port->adminPCost);
      }
#endif
      return STP_hop_2_state (this, SEND_STP);
    }

    if (port->owner->ForceVersion < 2) {
      port->adminPCost = computeAdjPCost (port, 0);
#ifdef STP_DBG
      if (port->pcost->debug) {
        stp_trace ("migrate adminPCost=%lu", port->adminPCost);
      }
#endif
      return STP_hop_2_state (this, SEND_STP);
    }

    break;
    case SEND_STP:
    return STP_hop_2_state (this, SENDING_STP);
    case SENDING_STP:
    if (port->mcheck) {
      port->adminPCost = computeAdjPCost (port, 2);
#ifdef STP_DBG
      if (port->pcost->debug) {
        stp_trace ("migrate adminPCost=%lu", port->adminPCost);
      }
#endif
      return STP_hop_2_state (this, SEND_RSTP);
    }
    if (port->mdelayWhile && (port->rcvdSTP || port->rcvdRSTP))
      return STP_hop_2_state (this, SENDING_STP);
    if (!port->mdelayWhile && port->rcvdRSTP) {
      port->adminPCost = computeAdjPCost (port, 2);
#ifdef STP_DBG
      if (port->pcost->debug) {
        stp_trace ("migrate adminPCost=%lu", port->adminPCost);
      }
#endif
      return STP_hop_2_state (this, SEND_RSTP);
    }
    break;
  }
  return False;
}
